dofile("../Include.lua")

Circuit = class(ScriptedShape)
Circuit.maxParentCount = -1
Circuit.maxChildCount = -1
Circuit.connectionInput =  sm.interactable.connectionType.logic
Circuit.connectionOutput = sm.interactable.connectionType.logic
Circuit.colorNormal = sm.color.new(0xB33229FF)
Circuit.colorHighlight = sm.color.new(0xEB584EFF)
Circuit.circuitSimulator = nil
Circuit.inputs = nil
Circuit.outputs = nil
Circuit.ready = nil
Circuit.gui = nil
Circuit.logger = nil
local scanner

function Circuit:init()
    ScriptedShape.init(self)
    self.ready = false
    self.logger = Logger(LogLevelEnum.Information, true)
end

local function extractSerialisedData(data)
    local tempLookup = Dictionary()
    local vertices = VertexSet()
    local edges = EdgeSet()
    local colourMap = Dictionary()

    for _, sGate in pairs(data.vertices) do
        local gate = Gate(sGate.identifier, sGate.booleanOperation)
        vertices:add(gate)
        tempLookup:add(gate:getIdentifier(), gate)
    end

    for _, sEdge in ipairs(data.edges) do
        local source = tempLookup:get(sEdge.source)
        local destination = tempLookup:get(sEdge.destination)
        local edge = Edge(source, destination)
        edges:add(edge)
    end

    for gateId, colour in pairs(data.colourMap) do
        colourMap:add(tonumber(gateId), colour)
    end

    return vertices, edges, colourMap
end

local function serialiseData(vertices, edges, colourMap)
    local data = {
        version = StorageManager.VERSION,
        vertices = vertices:toTable(),
        edges = edges:toTable(),
        colourMap = {}
    }

    for gateId, colour in colourMap:getIterator() do
        data.colourMap[tostring(gateId)] = colour
    end

    return data
end

local function prepareIO(self, colourMap)
    self.inputs = CircuitIO(self.circuitSimulator:getInputs(), colourMap, IOModeEnum.Input)
    self.outputs = CircuitIO(self.circuitSimulator:getOutputs(), colourMap, IOModeEnum.Output)
end

local function removeAllConnections(self)
    local scriptedShape
    for _, parent in ipairs(self.interactable:getParents()) do
        scriptedShape = sm.game.scriptedShapes:getShapeByShapeId(parent:getShape():getId())

        if scriptedShape ~= nil then
            local shapeUuid = tostring(scriptedShape.shape:getShapeUuid())

            if shapeUuid == "b2a2f556-d081-40d7-be2c-9a0291bb5c4d" or shapeUuid == "d1f4c1ba-b493-45da-bc74-cae285f2f8a3" then
                scriptedShape:server_resetConnectionState()
            end
        end

        parent:disconnect(self.interactable)
    end

    for _, child in ipairs(self.interactable:getChildren()) do
        scriptedShape = sm.game.scriptedShapes:getShapeByShapeId(child:getShape():getId())

        if scriptedShape ~= nil then
            local shapeUuid = tostring(scriptedShape.shape:getShapeUuid())

            if shapeUuid == "b2a2f556-d081-40d7-be2c-9a0291bb5c4d" or shapeUuid == "d1f4c1ba-b493-45da-bc74-cae285f2f8a3" then
                scriptedShape:server_resetConnectionState()
            end
        end

        self.interactable:disconnect(child)
    end
end

function Circuit:server_onCreate()
    ScriptedShape.server_onCreate(self)
    self.storageMgr = StorageManager(self, self.storage,"Circuits", "circuit.json")

    if self:server_hasStoredData() then
        self:server_createSimulator(extractSerialisedData(self.storageMgr:loadFromWorld()))
    end
end

function Circuit:server_onDestroy()
    ScriptedShape.server_onDestroy(self)
end

function Circuit:server_onFixedUpdate(deltaTime)
    ScriptedShape.server_onFixedUpdate(self, deltaTime)

    if self.ready then
        local outputSequence = self.circuitSimulator:simulate(self.inputs:getSequence())
        self.outputs:setSequence(outputSequence)
    end
end

function Circuit:server_addInput(shape, colour, pin)
    pin = pin or self.inputs:getAvailablePin(colour)

    if pin ~= nil and pin ~= "" then
        self.logger:log(LogLevelEnum.Trace, "[Server] Input added")
        self.inputs:attach(shape, pin)
        shape:server_setPin(pin)
    else
        shape.interactable:disconnect(self.interactable)
    end
end

function Circuit:server_removeInput(shape, pin)
    if self.inputs ~= nil then
        self.logger:log(LogLevelEnum.Trace, "[Server] Input removed")
        self.inputs:detach(shape, pin)
        shape:server_setPin(nil)
    end
end

function Circuit:server_addOutput(shape, colour, pin)
    pin = pin or self.outputs:getAvailablePin(colour)

    if pin ~= nil and pin ~= "" then
        self.logger:log(LogLevelEnum.Trace, "[Server] Output added")
        self.outputs:attach(shape, pin)
        shape:server_setPin(pin)
    else
        shape.interactable:disconnect(self.interactable)
    end
end

function Circuit:server_removeOutput(shape, pin)
    if self.outputs ~= nil then
        self.logger:log(LogLevelEnum.Trace, "[Server] Output removed")
        self.outputs:detach(shape, pin)
        shape:server_setPin(nil)
    end
end

function Circuit:server_createSimulator(vertices, edges, colourMap, player)
    local sortedGates, combinationalGates, sequentialGates, modifiedEdges = TopSorter.sort(vertices, edges:clone())
    self.circuitSimulator = CircuitSimulator(sortedGates, combinationalGates, sequentialGates, modifiedEdges)
    prepareIO(self, colourMap)
    self.storageMgr:saveToWorld(serialiseData(vertices, edges, colourMap))
    self.ready = true

    -- Log and client notification
    local circuitInfoMsg = { colour = "#917171", text = ("[Server] %i vertices and %i edges have been created"):format(vertices:getLength(), edges:getLength()) }
    self.logger:log(LogLevelEnum.Information, circuitInfoMsg.text )

    if player ~= nil then
        self.network:sendToClient(player, "client_alertCallback", { colour = "#C2C038", text = "[Server] Circuit created" })
        self.network:sendToClient(player, "client_printChatMessageCallback", circuitInfoMsg)
    end
end

function Circuit:server_beginScan(args, player)
    self.logger:log(LogLevelEnum.Trace, "[Server] Request host to scan")
    self.network:sendToClient(sm.player.getAllPlayers()[1], "host_scanCircuit", { player = player })
end

function Circuit:server_finishScan(args, player)
    removeAllConnections(self)
    self:server_createSimulator(scanner:getVertexSet(), scanner:getEdgeSet(), scanner:getColourMap(), args.player)
end

function Circuit:server_exportCircuit(args, player)
    self.logger:log(LogLevelEnum.Trace, "[Server] Export request received")
    local msg

    if self:server_hasStoredData() then
        if sm.isHost then
            self.storageMgr:saveToJsonFile(self.storageMgr:loadFromWorld())
            msg = { colour = "#C2C038", text = "[Server] Circuit saved on drive" }
            self.logger:log(LogLevelEnum.Information, msg.text)
        else
            msg = { colour = #C23838"", text = "[Server] Only the host can initiate exports" }
            self.logger:log(LogLevelEnum.Warning, msg.text)
        end
    else
        msg = { colour = "#C23838", text = "[Server] No circuit has been saved" }
        self.logger:log(LogLevelEnum.Warning, msg.text)
    end

    self.network:sendToClient(player, "client_alertCallback", msg)
end

function Circuit:server_importCircuit(args, player)
    self.logger:log(LogLevelEnum.Trace, "[Server] Import request received")
    local msg

    if sm.isHost then
        removeAllConnections(self)
        local config = self.storageMgr:loadFromJsonFile()
        local vertices, edges, colourMap = extractSerialisedData(config)
        self:server_createSimulator(vertices, edges, colourMap, player)
    else
        msg = { colour = "#C23838", text = "[Server] Only the host can initiate imports" }
        self.logger:log(LogLevelEnum.Warning, msg.text)
        self.network:sendToClient(player, "client_alertCallback", msg)
    end
end

function Circuit:client_onCreate()
    self.gui = sm.gui.createGuiFromLayout("$MOD_DATA/Gui/CircuitDialogue.layout")
    self.gui:setButtonCallback("Scan", "gui_scanCallback")
    self.gui:setButtonCallback("Export", "gui_exportCallback")
    self.gui:setButtonCallback("Import", "gui_importCallback")
    self.gui:setOnCloseCallback("gui_onCloseCallback")
end

function Circuit:client_onDestroy()
    self.gui:destroy()
end

function Circuit:gui_scanCallback(buttonName)
    self.network:sendToServer("server_beginScan")
    self.logger:log(LogLevelEnum.Trace, "[Client] Scan request sent")
    self.gui:close()
end

function Circuit:gui_exportCallback(buttonName)
    self.network:sendToServer("server_exportCircuit")
    self.logger:log(LogLevelEnum.Trace, "[Client] Export request sent")
    self.gui:close()
end

function Circuit:gui_importCallback(buttonName)
    self.network:sendToServer("server_importCircuit")
    self.logger:log(LogLevelEnum.Trace, "[Client] Import request sent")
    self.gui:close()
end

function Circuit:gui_onCloseCallback()
    self.gui:close()
end

function Circuit:client_canInteract()
    sm.gui.setInteractionText("Use", sm.gui.getKeyBinding("Use"), "to open the circuit dialogue")

    return true
end

function Circuit:client_onInteract(character, lookingAt)
    if lookingAt then
        self.gui:open()
    end
end

function Circuit:host_scanCircuit(args)
    self.logger:log(LogLevelEnum.Trace, "[Host] Scan request received")
    local startVertices = List()

    for i, child in ipairs(self.interactable:getChildren()) do
        local shapeUuid = tostring(child:getShape():getShapeUuid())

        if shapeUuid == "9f0f56e8-2c31-4d83-996c-d00a9b296c3f"
        or shapeUuid == "8f98db04-72eb-4a3a-88a1-f4f3e8d818ee" then
            if #child:getParents() > 1 then
                self.logger:log(LogLevelEnum.Warning, "[Host] Scanner is connected to at least one non input gate")

                return
            end

            startVertices:add(child)
        end
    end

    scanner = Scanner(startVertices)
    scanner:scan()
    self.logger:log(LogLevelEnum.Information, "[Host] Scan successful")
    self.network:sendToServer("server_finishScan", args)
end

function Circuit:client_printChatMessageCallback(msg)
    sm.gui.chatMessage(("%s%s"):format(msg.colour, msg.text))
end

function Circuit:client_alertCallback(msg)
    sm.gui.displayAlertText(("%s%s"):format(msg.colour, msg.text))
end
